var fs = require('fs');
var os = require('os');
var path = require('path');
var tsapi = require('typescript-api');
var cmpout = require('component-consoler');
var resolve = require('component-resolver');
var flatten = require('component-flatten');
var glob = require('./resolve-glob.js');
var dts = require('dts-bundle');

var maps_path = "./maps";

var red = '31';
var green = '32';
var yellow = '33';

function tsc(opts) {
    opts = opts || {};
    opts.install = opts.install || false;
    opts.development = opts.development || true;
    opts.verbose = true;

    // resolve the dependency tree
    resolve(process.cwd(), opts, function (err, tree) {
        if (err) throw err;

        var branches = flatten(tree);

        clean();
        
        generate_dependency_definitions(tree, branches);

        compile_scripts(tree, branches);
    });
}

function clean() {
	// clean the maps folder?
}	

function compile_scripts(tree, branches) {
	// No scripts field in the component.json file
    if (!tree.node.scripts)
     	return;

	// var tsscripts = glob(tree.node.scripts.map(function(jspath) { return change_extension(jspath, '.ts'); }));
    // compile all ".ts" files underneath the main path
    var tsscripts = glob(path.dirname(tree.node.main) + "/**/*.ts");
    var settings = get_compiler_settings(tree);

    var dtsfiles = branches.filter(function(branch) {
            return is_typecript_dependency(branch);
        })
        .map(function(branch) {
            return path.join(branch.path, get_definition(branch.node)); 
        });
	
	var batchCompiler = new tsapi.BatchCompiler(tsapi.IO);
    batchCompiler.logger = new tsapi.NullLogger();
    batchCompiler.compilationSettings = settings;
    batchCompiler.resolvedFiles = [];	
    batchCompiler.inputFiles = tsscripts.concat(dtsfiles);
    
    batchCompiler.addDiagnostic = function(diag) {
    	output_diagnostic(diag, batchCompiler);
    };

    batchCompiler.tryWriteOutputFiles = function(outputFiles) {
        if (!write_output_files(outputFiles, tree.node.main)) {
            batchCompiler.hasErrors = true;
            return false;
        }

        return true;
    }

    batchCompiler.resolve();

    if (batchCompiler.hasErrors) {
    	cmpout.log('typescript', 'resolution failed...', red);
    }
    else {
		batchCompiler.compile();

	    if (batchCompiler.hasErrors) {
    		cmpout.log('typescript', 'compilation failed...', red);
    	}
    }

    generate_bundle_definitions(tree, branches, batchCompiler);       

    if (!batchCompiler.hasErrors) {
    	cmpout.log('typescript', 'compilation succeeded', green);
    }
}

function get_compiler_settings(tree) {
	var settings = new tsapi.CompilationSettings();

	settings.codeGenTarget = "es5";
	settings.moduleGenTarget = "commonjs";
	settings.useCaseSensitiveFileResolution = false;
	settings.mapSourceFiles = true;
	settings.noImplicitAny = false;
	settings.noLib = false;
    settings.noExternalResolve = true;
    settings.outDirOption = maps_path;
    //settings.mapRoot = './maps';

    if (get_definition(tree.node))
       settings.generateDeclarationFiles = true;

	return tsapi.ImmutableCompilationSettings.fromCompilationSettings(settings);
}

function write_output_files(outputFiles, main) {
    try {
        outputFiles.forEach( function(outputFile) {
            if( ends_with(outputFile.name, '.d.ts')) {
                //var dtsname = path.join('./maps', path.relative(path.dirname(main), outputFile.name));
                //ensure_directory(dtsname);
                //fs.writeFileSync(dtsname, outputFile.text, 'utf8');
                ensure_directory(outputFile.name);
                fs.writeFileSync(outputFile.name, outputFile.text, 'utf8');
            }
            else if( ends_with(outputFile.name, '.js.map')) {
                //var mapname = path.join('./maps', path.relative(path.dirname(main), outputFile.name));
                //ensure_directory(mapname);
                //fs.writeFileSync(mapname, outputFile.text, 'utf8');
                ensure_directory(outputFile.name);
                fs.writeFileSync(outputFile.name, outputFile.text, 'utf8');
            }
            else if( ends_with(outputFile.name, '.js')) {
                var relname = path.relative(maps_path, outputFile.name);                        
                var jsname = path.join(path.dirname(main), relname);
                fs.writeFileSync(jsname, outputFile.text, 'utf8');
            }
        });
    } catch (e) {
        cmpout.log('typescript', "error creating '" + e.path + "' [" + e.code + "]", red);
        return false;
    }

    return true;
}

/*
 * Creates a file called _dependencies.d.ts alongside your component's main.
 * It contains reference to each of the dependencies typescript definitions.
 */
function generate_dependency_definitions(tree, branches) {
    var header = "/// ------------------------------------- ///" + os.EOL;
    header    += "/// Automatically generated. DO NOT EDIT. ///" + os.EOL;
    header    += "/// ------------------------------------- ///" + os.EOL + os.EOL;

   var content = branches
        .filter(function(branch) {
            return is_typecript_dependency(branch);
        })
        .reduce(function(current, branch) {
            var dtsfile = path.join(branch.path, get_definition(branch.node)); 

            current += '/// <reference path="';
            current += forward_slash( path.relative(path.dirname(tree.node.main), dtsfile) );
            current += '" />' + os.EOL;
    
            return current;
        }, header);

    // create a _dependencies.d.ts file in the same location as
    // the main file for this component    
    var filename = path.join(path.dirname(tree.node.main), "_dependencies.d.ts");
    fs.writeFileSync(filename, content);
    cmpout.log('generated', forward_slash(filename));
}

/*
 * If your component.json has a "typescript" file we create an 
 * external definition file for your component and place it there.
 */
function generate_bundle_definitions(tree, branches, compiler) {
    var outputFile = get_definition(tree.node);

    if (outputFile) {
        var main = path.join(maps_path, path.basename(tree.node.main));
        dts.bundle({
            name: tree.node.name,
            main: change_extension(main, ".d.ts"),
            out: outputFile,
            baseDir: './'
        });

        var loosedefs = compiler.inputFiles.reduce( function(current, file) {
            var source = compiler.getSourceFile(file);

            if (starts_with(path.dirname(file), path.dirname(tree.node.main))) {
                if ( ends_with(file,  ".d.ts")) {
                    var defstext = strip_triple_comments(source.scriptSnapshot.text);

                    if (defstext.length > 0)
                        current += defstext + os.EOL; 
                }
            }

            return current;
        }, "");

        if (loosedefs.length > 0)
            fs.appendFileSync(outputFile, loosedefs);

        cmpout.log('generated', outputFile);
    }
    else {
        cmpout.log('typescript', 'no definition file generated', yellow);   
    }
}

function strip_triple_comments(text) {
    var started = false;

    return text
        .replace(/\r\n/gm, "\n")
        .split("\n")        
        .reduce( function(current, line) {
            if (!starts_with(line, "///")) {  
                if ((line.length > 0) || started ) {
                    started = true;
                    current += line + os.EOL;
                }
            }
            return current;
        }, "");
}

function forward_slash(str) {
    return str.replace(/\\/g, '/');
}

function is_typecript_dependency(branch) {
    return (get_definition(branch.node) && (branch.type != 'local'))
}

function get_definition(node) {
    if( node.typescript ) {
        if (typeof node.typescript == 'string' || node.typescript instanceof String)
            return node.typescript;
        else
            return node.typescript.definition;                    
    }  
    return undefined;
}

function change_extension(file, ext) {
    return path.join(path.dirname(file), path.basename(file, path.extname(file)) + ext);
}

function ensure_directory(filename) {
    var dirname = path.dirname(filename);

    if( !fs.existsSync(dirname)) {
        ensure_directory(dirname);
        fs.mkdirSync(dirname, 0766)
    }
}

function starts_with(str, pref) {
    return (str.indexOf(pref) === 0);
}

function ends_with(str, post) {
    return str.lastIndexOf(post) + post.length === str.length;
}

function output_diagnostic(diag, compiler) {
    var color = yellow;

    if (diag.info().category === 1 /* Error */) {
        color = red;
        compiler.hasErrors = true;
    }

    if (diag.fileName()) {
        cmpout.log('typescript', tsapi.TypeScriptCompiler.getLocationText(diag, function(abs) { return path.relative("./", abs); } ), color);
    }

    cmpout.log('typescript', diag.message(), color);
}

module.exports = tsc;